@@ -7,6 +7,7 @@ import random
 import time
 import typing
 
+import aiohttp
 import fastapi
 import unimatrix.runtime
 from fastapi import Request
@@ -116,12 +117,7 @@ class Application(fastapi.FastAPI):
             # If there is an URL configuration file and it exists, add the routes
             # from it to the application.
             if self.urlconf is not None:
-                if not os.path.exists(self.urlconf):
-                    self.logger.warning(
-                        "URL configuration file does not exist: %s", self.urlconf
-                    )
-                else:
-                    urlconf.fromfile(self, self.urlconf)
+                self.setup_routes(self.urlconf)
 
 
         @self.on_event('shutdown')
@@ -232,13 +228,14 @@ class Application(fastapi.FastAPI):
                 request, TryAgain(30)
             )
         elif isinstance(exception, CanonicalException):
-            if exception.http_status_code >= 500:
+            status_code = getattr(exception, 'http_status_code', None) or 500
+            if status_code >= 500:
                 exception.log(self.logger.exception)
             response = JSONResponse(
-                status_code=exception.http_status_code,
-                content=exception.as_dict()
+                status_code=status_code,
+                content=exception.as_dict(),
+                headers=exception.http_headers
             )
-            response.headers['X-Canonical-Exception'] = type(exception).__name__
             return response
         else:
             raise NotImplementedError
@@ -269,3 +266,11 @@ class Application(fastapi.FastAPI):
     def get_issuer(self, request: Request) -> str:
         """Return the issuer used for JWS access tokens."""
         return self.issuer or f'{request.url.scheme}://{request.url.netloc}'
+
+    def setup_routes(self, path: str) -> None:
+        if not os.path.exists(path):
+            self.logger.warning(
+                "URL configuration file does not exist: %s", path
+            )
+        else:
+            urlconf.fromfile(self, path)
