@@ -92,3 +92,9 @@ class TryAgain(CanonicalException):
 
     def __init__(self, retry_after): # pragma: no cover
         super().__init__()
+
+
+class ParseError(CanonicalException):
+    http_status_code = 415
+    code = "UNSUPPORTED_MEDIA_TYPE"
+    message = "The server could not process the request body."
