@@ -10,26 +10,35 @@ import urllib.parse
 from collections import OrderedDict
 
 import ioc
+import fastapi
 from fastapi import APIRouter
 from fastapi import Depends
 from fastapi import Request
+from fastapi.exceptions import RequestValidationError
 from fastapi.responses import Response
 from fastapi.responses import RedirectResponse
 from fastapi.security import HTTPBearer
 from fastapi.security import HTTPAuthorizationCredentials
+from fastapi.dependencies.utils import get_dependant
+from fastapi.dependencies.utils import get_parameterless_sub_dependant
+from fastapi.dependencies.utils import solve_dependencies
 from pydantic import BaseModel
 from unimatrix.conf import settings
 from unimatrix.lib.datastructures import ImmutableDTO, DTO
 
 from .auth import IHTTPAuthenticationService
 from .auth import IPrincipal
+from .bodyconsumer import BodyConsumer
 from .dependency import inject
 from .exceptions import BearerAuthenticationRequired
 from .endpointiam import EndpointIAM
 from .exceptions import NotAuthorized
 from .exceptions import TrustIssues
+from .parsers import IParser
 from .requestprincipalresolver import RequestPrincipalResolver
 from .resourceschema import ResourceSchema
+from .utils import as_dependant
+from .utils import class_dependency
 from .utils import clean_signature
 from .utils import clone_signature
 
@@ -38,6 +47,7 @@ ACTION_METHODS = {
     'apply'     : 'PATCH',
     'create'    : 'POST',
     'destroy'   : 'DELETE',
+    'exists'    : 'HEAD',
     'index'     : 'GET',
     'purge'     : 'DELETE',
     'replace'   : 'PUT',
@@ -47,10 +57,18 @@ ACTION_METHODS = {
 
 DEFAULT_ACTIONS = [
     'create', 'retrieve', 'update', 'destroy', 'index', 'apply', 'replace',
-    'purge'
+    'purge', 'exists'
 ]
 
-DETAIL_ACTIONS = {'retrieve', 'update', 'replace', 'destroy', 'apply'}
+DETAIL_ACTIONS = {'retrieve', 'update', 'replace', 'destroy', 'apply', 'exists'}
+
+
+def wrap_exists(func):
+    @functools.wraps(func)
+    async def exists(*args, **kwargs):
+        result = await func(*args, **kwargs)
+        return Response(status_code=200 if result else 404)
+    return exists
 
 
 def get_return_type(handler):
@@ -64,7 +82,7 @@ class ResourceEndpointSetMetaclass(type):
 
     def __new__(cls, name, bases, attrs):
         super_new = super().__new__
-        if name in ('ResourceEndpointSet', 'PublicResourceEndpointSet'):
+        if name in ('BaseResourceEndpointSet', 'ResourceEndpointSet', 'PublicResourceEndpointSet'):
             return super_new(cls, name, bases, attrs)
 
         # Instantiate the IAM handler for the endpoint.
@@ -76,6 +94,9 @@ class ResourceEndpointSetMetaclass(type):
             local=attrs.pop('trust_local', False)
         )
 
+        if 'exists' in attrs:
+            attrs['exists'] = wrap_exists(attrs['exists'])
+
         new_class = super_new(cls, name, bases, attrs)
         hints = typing.get_type_hints(new_class)
 
@@ -151,10 +172,11 @@ class ResourceEndpointSetMetaclass(type):
         return new_class
 
 
-class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
+class BaseResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
     """Groups a set of endpoints that allow reading, mutating and
     destroying a specific resource.
     """
+
     parent_class = None
 
     #: The list of valid action names.
@@ -226,6 +248,11 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
     #: The class used to handle IAM.
     iam_class: type = EndpointIAM
 
+    @property
+    def handler(self):
+        """Return the current request handler."""
+        return getattr(self, self.action)
+
     @functools.cached_property
     def parameters(self) -> typing.List[inspect.Parameter]:
         """Return the list of path parameters as :class:`inspect.Parameter`
@@ -318,6 +345,7 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
                 cls._create_annotated_handler(
                     cls, obj.action.name, obj, parent=parent,
                     detail=obj.action.detail,
+                    path=path,
                     path_parameters=path_parameters,
                     logger=router.logger
                 ),
@@ -375,6 +403,7 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
                     cls, action, handler,
                     parent=parent,
                     detail=is_detail,
+                    path=path,
                     path_parameters=path_parameters,
                     logger=router.logger
                 ),
@@ -457,6 +486,11 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
             inspect.Parameter.POSITIONAL_OR_KEYWORD,
             annotation=Request
         )
+        annotations['response'] = inspect.Parameter(
+            'response',
+            inspect.Parameter.POSITIONAL_OR_KEYWORD,
+            annotation=Response
+        )
 
         if detail and path_parameters:
             for param in path_parameters:
@@ -474,6 +508,8 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
         for varname, param in list(signature.parameters.items()):
             if param.kind == inspect.Parameter.VAR_POSITIONAL: # pragma: no cover
                 break
+            if isinstance(param.default, fastapi.params.Depends):
+                continue
             annotations[varname] = param
 
         # Remove the self parameter (causes incorrect rendering of the UI)
@@ -485,32 +521,16 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
             if param.kind != inspect.Parameter.VAR_POSITIONAL\
             and param.kind != inspect.Parameter.VAR_KEYWORD:
                 continue
+            if isinstance(param.default, fastapi.params.Depends):
+                continue
             annotations[varname] = param
 
-        # Setup authentication by injecting the Endpoint.get_principal()
-        # and Endpoint.authorize() results.
-        async def authorize(*args, **kwargs):
-            return await view_class.authorize(*args, **kwargs)
-        clone_signature(view_class.authorize, authorize, {
-            'principal': inspect.Parameter(
-                name='principal',
-                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
-                default=Depends(view_class.get_principal)
+        if view_class.require_authentication:
+            annotations['bearer'] = inspect.Parameter(
+                name='bearer',
+                kind=inspect.Parameter.KEYWORD_ONLY,
+                default=Depends(HTTPBearer(auto_error=False))
             )
-        })
-
-        annotations.update({
-            '_authorized': inspect.Parameter(
-                '_authorized',
-                inspect.Parameter.POSITIONAL_OR_KEYWORD,
-                default=Depends(authorize)
-            ),
-            '_principal': inspect.Parameter(
-                '_principal',
-                inspect.Parameter.POSITIONAL_OR_KEYWORD,
-                default=Depends(view_class.get_principal)
-            ),
-        })
 
         # Update the signature to hold the new annotations. Remove the self
         # parameter to not confuse FastAPI
@@ -524,10 +544,11 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
         return signature, annotations
 
     @staticmethod # pylint: disable=line-too-long
-    def _create_annotated_handler(view_class, action, handler, parent, detail, path_parameters, logger=None):
+    def _create_annotated_handler(view_class, action, handler, parent, detail, path, path_parameters, logger=None):
         @functools.wraps(handler)
         async def request_handler(
             request: Request,
+            response: Response,
             *args, **kwargs
         ):
             """Wrapper function that ensures that the proper dependencies
@@ -535,9 +556,9 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
             """
             view = view_class(action)
             view.request = request
+            view.response = response
             view.logger = logger
-            view.principal = kwargs.pop('_principal', None)
-            view.is_authorized = kwargs.pop('_authorized', False)
+            view.path_format = path
             return await view.dispatch(*args, **kwargs)
 
         request_handler.__signature__, request_handler.__annotations__ = (
@@ -550,27 +571,89 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
         )
         return request_handler
 
+    def get_permissions(self) -> set:
+        """Return the set of permissions that are required to perform the
+        operation the :class:`Endpoint` is currently handling.
+        """
+        return getattr(self.handler, 'permissions', set())
+
+    def set_cookie(self,
+        key: str,
+        value: str,
+        max_age: int = None,
+        expires: int = None,
+        path: str = None,
+        domain: str = None,
+        secure: bool = None,
+        httponly: bool = False,
+        samesite: str = 'lax',
+    ) -> None:
+        """Sets a cookie on the response."""
+        return self.response.set_cookie(
+            key=key,
+            value=value,
+            max_age=max_age,
+            expires=expires,
+            path=path,
+            domain=domain,
+            secure=secure,
+            httponly=httponly,
+            samesite=samesite
+        )
+
     async def dispatch(self, *args, **kwargs) -> dict:
         """Dispatch the incoming HTTP request to the appropriate request
         handler.
         """
+        body = await self.get_body()
+        current_user = class_dependency(self.get_principal, 'principal')
+        handler = as_dependant(getattr(self, self.action))
+        dependant = get_dependant(path=self.path_format, call=handler)
+        dependant.dependencies.extend([
+            get_parameterless_sub_dependant(
+               depends=current_user,
+               path=self.path_format
+            ),
+            get_parameterless_sub_dependant(
+                depends=class_dependency(
+                    self.authorize,
+                    dst='is_authorized',
+                    principal=current_user
+                ),
+                path=self.path_format
+            )
+        ])
+        values, errors, tasks, response, _ =  await solve_dependencies(
+            request=self.request,
+            dependant=dependant,
+            body=body,
+            dependency_overrides_provider=None
+        )
+        if errors:
+            raise RequestValidationError(errors, body=None)
         if self.principal is None and self.require_authentication:
             raise BearerAuthenticationRequired
         if not self.is_authorized:
             raise NotAuthorized
-        handler = getattr(self, self.action)
-        assert callable(handler), handler #nosec
-        return await handler(
-            *args,
-            **(await self.preprocess_params(**kwargs))
+        if response:
+            if response.status_code is not None:
+                self.response.status_code = response.status_code
+            if response.headers:
+                self.response.headers.update(response.headers)
+        result = await dependant.call(
+            **await self.preprocess_params(**values)
         )
+        return result
 
     async def preprocess_params(self, **params) -> dict:
         """Hook to modify the parameters passed to a request handler."""
         return params
 
-    @classmethod
-    async def get_principal(cls,
+    async def authorize(self, *args, **kwargs):
+        assert hasattr(self, 'principal') # nosec
+        return True
+
+    async def get_principal(self,
         request: Request,
         bearer: HTTPAuthorizationCredentials = Depends(HTTPBearer(auto_error=False)),
         resolver = inject(
@@ -578,26 +661,17 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
             default=RequestPrincipalResolver()
         )
     ) -> IPrincipal:
-        """Return a :class:`unimatrix.ext.webapi.auth.IPrincipal` implementation
-        based on the current request and supplied credentials, or ``None`` if
-        the credentials were absent or did not resolve to a principal.
-        """
         if bearer is None:
+            if self.require_authentication:
+                raise BearerAuthenticationRequired
             return None
         return await resolver.resolve(
             request=request,
-            iam=cls.iam,
+            iam=self.iam,
             bearer=bearer.credentials,
-            policy=cls.trust_policy
+            policy=self.trust_policy
         )
 
-    @staticmethod
-    async def authorize(principal: IPrincipal, *args, **kwargs) -> bool:
-        """Hook to perform authorization. Implementations are expected to raise
-        an exception if the request is not authorized to perform the operation.
-        """
-        return True
-
     def render_to_response(self, *args, **kwargs) -> Response:
         """Renders a resource to a HTTP response."""
         return self.resource_class.render_to_response(
@@ -713,6 +787,10 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
         return params
 
 
+class ResourceEndpointSet(BaseResourceEndpointSet, BodyConsumer):
+    pass
+
+
 class PublicResourceEndpointSet(ResourceEndpointSet):
     """A :class:`ResourceEndpointSet` implementation that is public i.e.
     no authentication or authorization is performed.
